/*------------------------------------------------------------------------------

Copyright 2012 University of Aveiro, Portugal, All Rights Reserved.

These programs are supplied free of charge for research purposes only,
and may not be sold or incorporated into any commercial product. There is
ABSOLUTELY NO WARRANTY of any sort, nor any undertaking that they are
fit for ANY PURPOSE WHATSOEVER. Use them at your own risk. If you do
happen to find a bug, or have modifications to suggest, please report
the same to Armando J. Pinho, ap@ua.pt. The copyright notice above
and this statement of conditions must remain an integral part of each
and every copy made of these files.

------------------------------------------------------------------------------*/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include <limits.h>
#include <float.h>
#include <ctype.h>
#include <time.h>

#include "defs.h"
#include "mem.h"
#include "seqs.h"
#include "context.h"

#define	MISSING_BASE	-1

#define UPDATE_CMODELS	1

#define MODEL_STATIC	0
#define MODEL_DYNAMIC	1
#define SELF_MIN		2
#define SELF_LR			3
#define SELF_RL			4

#define W_HAMMING       0
#define W_HANN          1
#define W_BLACKMAN      2
#define W_RECTANGULAR   3

#define DEFAULT_MODEL_ARGS "comprof -k 3 -k 7 -k 11 -k 14 1/20"

double (* Power)(double x, double y) = Pow;

char *helpMsg = "\n\
This program generates information profiles of DNA sequences using\n\
combinations of finite-context (Markov) models. It has several modes\n\
of operation, namely:\n\n\
- Model static (ms): a sequence (the model sequence) is used to load\n\
the statistics into the finite-context models. The profile of the target\n\
sequence is computed keeping the models frozen. If the model sequence is\n\
not provided (using the -ms option), then the target sequence is used\n\
for modeling.\n\n\
- Model dynamic (md): a sequence (the model sequence) is used to load\n\
the statistics into the finite-context models. When the profile of the\n\
target sequence is computed, the models continue to be updated. If the\n\
model sequence is not provided (using the -ms option), then the target\n\
sequence is used for modeling.\n\n\
- Self minimum (min): the target sequence is processed in both directions\n\
and the profile is obtained by picking up the minimum value at each\n\
position. The models are continuously updated.\n\n\
- Left to right (lr): the target sequence is processed from left to right.\n\n\
- Right to left (rl): the target sequence is processed from right to left.\n\n\
The profile can be low-pass filtered and sub-sampled. The -w option allows\n\
the definition of a filtering window of 2M+1 samples. Possible window\n\
shapes are: Hamming (-hm option), Hann (-hn option), Blackman (-bk option)\n\
and Rectangular (-rc option). If the -s option is defined, a sub-sampling\n\
of 1:S is applied, i.e. (S-1) out of S samples are discarded.\n";

/*----------------------------------------------------------------------------*/

double Filter(float *profile, int size, int n, int M, double *w)

	{
	int k;
	double sum = 0, wSum = 0;

	for(k = -M ; k <= M ; k++)
		if(n + k >= 0 && n + k < size && profile[n+k] != MISSING_BASE)
			{
			sum += w[M+k] * profile[n+k];
			wSum += w[M+k];
			}

	return wSum ? sum / wSum : MISSING_BASE;
	}

/*----------------------------------------------------------------------------*/

void ParseListOfPositions(char *str, CTemplate *cTemplate)

	{
	int n, low, high;

	cTemplate->size = 0;
	while(1)
		{
		switch(sscanf(str, "%d:%d", &low, &high))
			{
			case 1:
				if(low == 0)
					return;

				if(low < 0)
					{
					fprintf(stderr,"Error: can't handle non-causal contexts\n");
					exit(1);
					}

				IncreaseTemplateStorage(cTemplate, 1);
				cTemplate->position[cTemplate->size++] = low;
				break;

			case 2:
				if(low <= 0)
					{
					fprintf(stderr,"Error: can't handle non-causal contexts\n");
					exit(1);
					}

				if(high - low + 1 > 0)
					{
					IncreaseTemplateStorage(cTemplate, high - low + 1);
					for(n = 0 ; n < high - low + 1 ; n++)
						cTemplate->position[cTemplate->size++] = low + n;

					}

				break;

			default:
				return;

			}

		while(*str != '\0' && *str++ != ',')
			;

		}

	}

/*----------------------------------------------------------------------------*/

void Run(Sequence *seq, Alphabet *alphabet, CModel **cModels, PModel **pModels,
  CTemplate *cTemplate, int nCModels, int updateCModels, float *profile,
  unsigned seqStart, unsigned seqEnd)

	{
	int n, cModel;
	double *cModelWeight, cModelTotalWeight = 0;
	unsigned pModelIdx;
	Symbol s, symbol, symbolC;
	PModel *mxPModel = CreatePModel(N_SYMBOLS);
	FloatPModel *floatPModel = CreateFloatPModel(N_SYMBOLS);

	/*
	 * The weight associated with each cModel in the mixture model.
	 */
	if(!(cModelWeight = (double *)Calloc(nCModels, sizeof(double))))
		{
		fprintf(stderr, "Error: out of memory\n");
		exit(1);
		}

	for(cModel = 0 ; cModel < nCModels ; cModel++)
		{
		/* Assume an uniform prior */
		cModelWeight[cModel] = 1.0 / nCModels;
		cModelTotalWeight += cModelWeight[cModel];
		}

	for(n = seqStart - 1 ; n < seqEnd ; n++)
		{
		if((n - seqStart - 1) % ((seqEnd - seqStart) / 100 + 1) == 0)
			fprintf(stderr, "%3d%%\r", (int)((n - seqStart - 1) /
			  ((seqEnd - seqStart) / 100 + 1)));

		if((symbol = alphabet->symbols[seq->bases[n]]) > 3)
			{
			/* Base does not belong to ACGT or acgt */
			if(profile)
				profile[n] = MISSING_BASE;

			continue;
			}

		for(s = 0 ; s < N_SYMBOLS ; s++)
			floatPModel->freqs[s] = 0;

		for(cModel = 0 ; cModel < nCModels ; cModel++)
			{
			pModelIdx = GetPModelIdx(seq->bases + n,
			  &cTemplate[cModel], alphabet);
			ComputePModel(cModels[cModel], pModels[cModel], pModelIdx);

			/* Re-normalize the weights */
			cModelWeight[cModel] /= cModelTotalWeight;

			/*
			 * The probabilities estimated by each cModel are weighted
			 * according to the set of current weights.
			 */
			for(s = 0 ; s < N_SYMBOLS ; s++)
				floatPModel->freqs[s] += (double)pModels[cModel]->freqs[s] /
				  pModels[cModel]->sum * cModelWeight[cModel];

			}

		mxPModel->sum = 0;
		for(s = 0 ; s < N_SYMBOLS ; s++)
			{
			mxPModel->freqs[s] = 1 +
			  (unsigned)(floatPModel->freqs[s] * PMODEL_MAX_COUNT);
			mxPModel->sum += mxPModel->freqs[s];
			}

		if(profile)
			profile[n] = PModelSymbolNats(mxPModel, symbol) / M_LN2;

		cModelTotalWeight = 0;
		for(cModel = 0 ; cModel < nCModels ; cModel++)
			{
			/*
			 * Update the "accumulated" probability generated by the
			 * model along the sequence.
			 */
			cModelWeight[cModel] =
			  Power(cModelWeight[cModel], cModels[cModel]->gamma) *
			  (double)pModels[cModel]->freqs[symbol] / pModels[cModel]->sum;

			cModelTotalWeight += cModelWeight[cModel];

			if(updateCModels)
				{
				/* Update the fcm counters */
				pModelIdx = GetPModelIdx(seq->bases + n,
				  &cTemplate[cModel], alphabet);
				UpdateCModelCounter(cModels[cModel], pModelIdx, symbol);
				if(cModels[cModel]->ir == 'y')
					{
					pModelIdx = GetPModelIdxC(seq->bases + n,
					  &cTemplate[cModel], alphabet);
					if((symbolC = alphabet->symbols[seq->
					  bases[n - cTemplate[cModel].deepestPosition]]) <= 3)
						/* Only if base belongs to ACGT or acgt */
						{
						symbolC = 3 - symbolC; /* Complement */
						UpdateCModelCounter(cModels[cModel], pModelIdx,symbolC);
						}

					}

				}

			}

		}

	}

/*----------------------------------------------------------------------------*/

void ReverseProfile(float *profile, int size)

	{
	int n;
	float tmp;

	for(n = 0 ; n < size / 2 ; n++)
		{
		tmp = profile[n];
		profile[n] = profile[size - n - 1];
		profile[size - n - 1] = tmp;
		}

	}

/*----------------------------------------------------------------------------*/

int StrToArgv(char *str, char ***argv)

	{
	int n, argc = 1;
	char *dup = strdup(str);

	n = 0;
	while(dup[n])
		if(dup[n++] == ' ')
			argc++;

	if(!(*argv = (char **)Malloc(argc * sizeof(char *))))
		{
		fprintf(stderr, "Error: in memory allocation\n");
		exit(1);
		}

	n = 0;
	(*argv)[0] = dup;
	argc = 1;
	do
		{
		if(dup[n] == ' ')
			{
			dup[n] = '\0';
			(*argv)[argc++] = dup + n + 1;
			}

		}
	while(dup[++n]);

	return argc;
	}

/*----------------------------------------------------------------------------*/

int main(int argc, char *argv[])

	{
	char verbose, omitIndexCol, updateInvComp, plot, *wigChrom = NULL, **mArgv;
	double gamma, *w;
	int n, k, nCModels, cModel, deepestPosition, mode, M, mArgc;
	FILE *fpOut = NULL, *fpPlot;
	unsigned hSize = 9999991; // hSizes {9999991,19999999,29999999,39999983}
	unsigned deltaNum, deltaDen, start, end, modelStart, modelEnd,
	  targetStart, targetEnd, outputStart, outputEnd, S, wType;
	CModel **cModels;
	CTemplate *cTemplate;
	char template[16];
	PModel **pModels;
	Sequence *modelSeq = NULL, *targetSeq;
	Alphabet *alphabet;
	float *profile, *tmpProfile, outValue;

	// Needed to measure CPU time
	clock_t tic, tac, clockStart;
	double cpuTimeUsed;

	clockStart = clock();

	verbose = 'n';
	plot = 'n';
	omitIndexCol = 'n';
	updateInvComp = 'y';
	gamma = 0.99;
	mode = SELF_MIN;
	M = -1;
	S = 1;
	wType = W_HAMMING;
	if(argc < 2)
		{
		fprintf(stderr, "Usage: comprof [ -h (help) ]\n");
		fprintf(stderr, "               [ -o outFile ]\n");
		fprintf(stderr, "               [ -wig chrom ]\n");
		fprintf(stderr, "               [ -v (print more info) ]\n");
		fprintf(stderr, "               [ -g gamma (def %.3f) ]\n", gamma);
		fprintf(stderr, "               [ -k order n/d ]\n");
		fprintf(stderr, "               [ -k order n/d ]\n");
		fprintf(stderr, "               ...\n");
		fprintf(stderr, "               [ -ic (do not update inverted complements) ]\n");
		fprintf(stderr, "               [ -ms modelSeq ]\n");
		fprintf(stderr, "               [ -mr start:end (model range) ]\n");
		fprintf(stderr, "               [ -tr start:end (target range) ]\n");
		fprintf(stderr, "               [ -or start:end (output range) ]\n");
		fprintf(stderr, "               [ -pow (standard lib pow) ]\n");
		fprintf(stderr, "               [ -1 (omit index col) ]\n");
		fprintf(stderr, "               [ -m ms/md/min/lr/rl (def min) ]\n");
		fprintf(stderr, "               [ -s S (def 1; sub-sampling 1:S) ]\n");
		fprintf(stderr, "               [ -w M (def 0 or 5*S if -s present; window size 2M+1) ]\n");
		fprintf(stderr, "               [ -hm (Hamming, def) ]\n");
		fprintf(stderr, "               [ -hn (Hann) ]\n");
		fprintf(stderr, "               [ -bk (Blackman) ]\n");
		fprintf(stderr, "               [ -rc (Rectangular) ]\n");
		fprintf(stderr, "               [ -p (gnuplot) ]\n");
		fprintf(stderr, "               targetSeq\n");
		return 1;
		}

	printf("comprof version 1.000\n");
	for(n = 1 ; n < argc ; n++)
		if(strcmp("-h", argv[n]) == 0)
			{
			fprintf(stderr, "%s\n", helpMsg);
			exit(0);
			}

	for(n = 1 ; n < argc ; n++)
		if(strcmp("-o", argv[n]) == 0)
			{
			if(!(fpOut = fopen(argv[n+1], "w")))
				{
				fprintf(stderr, "Error: can't open file\n");
				return 1;
				}

			break;
			}

	for(n = 1 ; n < argc ; n++)
		if(strcmp("-g", argv[n]) == 0)
			{
			gamma = atof(argv[n+1]);
			if(gamma >= 1 || gamma < 0)
				{
				fprintf(stderr, "Error: gamma should belong to [0, 1)\n");
				return 1;
				}

			break;
			}

	for(n = 1 ; n < argc ; n++)
		if(strcmp("-v", argv[n]) == 0)
			{
			verbose = 'y';
			break;
			}

	for(n = 1 ; n < argc ; n++)
		if(strcmp("-p", argv[n]) == 0)
			{
			plot = 'y';
			break;
			}

	for(n = 1 ; n < argc ; n++)
		if(strcmp("-pow", argv[n]) == 0)
			{
			Power = pow;
			break;
			}

	for(n = 1 ; n < argc ; n++)
		if(strcmp("-1", argv[n]) == 0)
			{
			omitIndexCol = 'y';
			break;
			}

	for(n = 1 ; n < argc ; n++)
		if(strcmp("-wig", argv[n]) == 0)
			{
			wigChrom = argv[n+1];
			break;
			}

	for(n = 1 ; n < argc ; n++)
		if(strcmp("-ic", argv[n]) == 0)
			{
			updateInvComp = 'n';
			break;
			}

	for(n = 1 ; n < argc ; n++)
		if(strcmp("-ms", argv[n]) == 0)
			{
			modelSeq = ReadSequence(argv[n+1]);
			printf("The model sequence has a total of %u bases\n",
			  modelSeq->size);
			if(verbose == 'y')
				for(n = 0 ; n < 256 ; n++)
					if(modelSeq->usage[n])
						printf("  %c : %u\n", n, modelSeq->usage[n]);

			break;
			}

	for(n = 1 ; n < argc ; n++)
		if(strcmp("-m", argv[n]) == 0)
			{
			if(strcmp("ms", argv[n+1]) == 0)
				mode = MODEL_STATIC;

			if(strcmp("md", argv[n+1]) == 0)
				mode = MODEL_DYNAMIC;

			if(strcmp("min", argv[n+1]) == 0)
				mode = SELF_MIN;

			if(strcmp("lr", argv[n+1]) == 0)
				mode = SELF_LR;

			if(strcmp("rl", argv[n+1]) == 0)
				mode = SELF_RL;

			break;
			}

	for(n = 1 ; n < argc ; n++)
		if(!strcmp("-w", argv[n]))
			{
			M = atoi(argv[n+1]);
			break;
			}

	for(n = 1 ; n < argc ; n++)
		if(!strcmp("-s", argv[n]))
			{
			S = atoi(argv[n+1]);
			if(S < 1)
				S = 1;

			break;
			}

	for(n = 1 ; n < argc ; n++)
		if(!strcmp("-hm", argv[n]))
			{
			wType = W_HAMMING;
			break;
			}

	for(n = 1 ; n < argc ; n++)
		if(!strcmp("-hn", argv[n]))
			{
			wType = W_HANN;
			break;
			}

	for(n = 1 ; n < argc ; n++)
		if(!strcmp("-bk", argv[n]))
			{
			wType = W_BLACKMAN;
			break;
			}

	for(n = 1 ; n < argc ; n++)
		if(!strcmp("-rc", argv[n]))
			{
			wType = W_RECTANGULAR;
			break;
			}

	targetSeq = ReadSequence(argv[argc-1]);
	printf("The target sequence has a total of %u bases\n", targetSeq->size);
	if(verbose == 'y')
		for(n = 0 ; n < 256 ; n++)
			if(targetSeq->usage[n])
				printf("  %c : %u\n", n, targetSeq->usage[n]);

	// Auto window size
	if(M == -1)
		{
		if(targetSeq->size > 50000)
			S = targetSeq->size / 50000;

		if(S == 1)
			M = 0;

		else
			M = S * 5;

		}

	// Allocate the filter coefficients and calculate them
	if(!(w = (double *)Malloc((2 * M + 1) * sizeof(double))))
		{
		fprintf(stderr, "Error: out of memory\n");
		return 1;
		}

	switch(wType)
		{
		case W_HAMMING:
			for(k = -M ; k <= M ; k++)
				w[M+k] = 0.54 + 0.46 * cos((2 * M_PI * k) / (2 * M + 1));

			break;

		case W_HANN:
			for(k = -M ; k <= M ; k++)
				w[M+k] = 0.5 * (1 + cos((2 * M_PI * k) / (2 * M + 1)));

			break;

		case W_BLACKMAN:
			for(k = -M ; k <= M ; k++)
				w[M+k] = 0.42 + 0.5 * cos((2 * M_PI * k) / (2 * M + 1)) +
				  0.08 * cos((4 * M_PI * k) / (2 * M + 1));

			break;

		case W_RECTANGULAR:
			for(k = -M ; k <= M ; k++)
				w[M+k] = 1;

			break;

		}

	// If the model sequence is not given, then use the target as model
	if(modelSeq == NULL)
		modelSeq = targetSeq;

	modelStart = 1;
	modelEnd = modelSeq->size;
	targetStart = 1;
	targetEnd = targetSeq->size;

	for(n = 1 ; n < argc ; n++)
		if(strcmp("-mr", argv[n]) == 0)
			{
			if(sscanf(argv[n+1], "%d:%d", &start, &end) == 2)
				{
				if(start > 0 && start <= modelSeq->size)
					modelStart = start;

				if(end > 0 && end <= modelSeq->size)
					modelEnd = end;

				break;
				}

			if(sscanf(argv[n+1], "%d:", &start) == 1)
				{
				if(start > 0 && start <= modelSeq->size)
					modelStart = start;

				break;
				}

			if(sscanf(argv[n+1], ":%d", &end) == 1)
				{
				if(end > 0 && end <= modelSeq->size)
					modelEnd = end;

				break;
				}

			break;
			}

	for(n = 1 ; n < argc ; n++)
		if(strcmp("-tr", argv[n]) == 0)
			{
			if(sscanf(argv[n+1], "%d:%d", &start, &end) == 2)
				{
				if(start > 0 && start <= targetSeq->size)
					targetStart = start;

				if(end > 0 && end <= targetSeq->size)
					targetEnd = end;

				break;
				}

			if(sscanf(argv[n+1], "%d:", &start) == 1)
				{
				if(start > 0 && start <= targetSeq->size)
					targetStart = start;

				break;
				}

			if(sscanf(argv[n+1], ":%d", &end) == 1)
				{
				if(end > 0 && end <= targetSeq->size)
					targetEnd = end;

				break;
				}

			break;
			}

	outputStart = targetStart;
	outputEnd = targetEnd;

	for(n = 1 ; n < argc ; n++)
		if(strcmp("-or", argv[n]) == 0)
			{
			if(sscanf(argv[n+1], "%d:%d", &start, &end) == 2)
				{
				if(start > 0 && start <= targetSeq->size)
					outputStart = start;

				if(end > 0 && end <= targetSeq->size)
					outputEnd = end;

				break;
				}

			if(sscanf(argv[n+1], "%d:", &start) == 1)
				{
				if(start >= targetStart && start <= targetEnd)
					outputStart = start;

				break;
				}

			if(sscanf(argv[n+1], ":%d", &end) == 1)
				{
				if(end >= targetStart && end <= targetEnd)
					outputEnd = end;

				break;
				}

			break;
			}

	// Compute the coding alphabet (symbols occurring in the target seq)
	if(!(alphabet = (Alphabet *)Calloc(1, sizeof(Alphabet))))
		{
		fprintf(stderr, "Error: out of memory\n");
		return 1;
		}

	// Consider 5 different symbols: a/A, c/C, g/G, t/T, all others
	AssignSymbols(alphabet);

	// Consider 4 different symbols: A, C, G, T, (N maps into A)
	AssignCtxSymbols(alphabet);

	if(!(profile = (float *)Malloc(targetSeq->size * sizeof(float))))
		{
		fprintf(stderr, "Error: out of memory\n");
		return 1;
		}

	nCModels = 0;
	for(n = 1 ; n < argc ; n++)
		if(strcmp("-k", argv[n]) == 0)
			nCModels++;

	if(nCModels == 0)
		{
		// If no models are defined in the command line, then use the default
		mArgc = StrToArgv(DEFAULT_MODEL_ARGS, &mArgv);
		nCModels = 0;
		for(n = 1 ; n < mArgc ; n++)
			if(strcmp("-k", mArgv[n]) == 0)
				nCModels++;

		}

	else
		{
		mArgv = argv;
		mArgc = argc;
		}

	if(!(cTemplate = (CTemplate *)Calloc(nCModels, sizeof(CTemplate))))
		{
		fprintf(stderr, "Error: out of memory\n");
		return 1;
		}

	if(!(cModels = (CModel **)Calloc(nCModels, sizeof(CModel *))))
		{
		fprintf(stderr, "Error: out of memory\n");
		return 1;
		}

	if(!(pModels = (PModel **)Calloc(nCModels, sizeof(PModel *))))
		{
		fprintf(stderr, "Error: out of memory\n");
		return 1;
		}

	cModel = 0;
	for(n = 1 ; n < mArgc ; n++)
		{
		if(strcmp("-k", mArgv[n]) == 0)
			{
			sprintf(template, "1:%s", mArgv[n+1]);
			ParseListOfPositions(template, &cTemplate[cModel++]);
			}

		}

	tic = clock();
	cpuTimeUsed = ((double) (tic - clockStart)) / CLOCKS_PER_SEC;
    printf("Needed %g s for reading the sequence(s)\n", cpuTimeUsed);

	deepestPosition = 0;
	for(cModel = 0 ; cModel < nCModels ; cModel++)
		{
		if(cTemplate[cModel].size == 0)
			continue;

		cTemplate[cModel].deepestPosition = cTemplate[cModel].position[0];
		for(n = 1 ; n < cTemplate[cModel].size ; n++)
			if(cTemplate[cModel].position[n] >
			  cTemplate[cModel].deepestPosition)
				cTemplate[cModel].deepestPosition =
				  cTemplate[cModel].position[n];

		if(deepestPosition < cTemplate[cModel].deepestPosition)
			deepestPosition = cTemplate[cModel].deepestPosition;

		}

	cModel = 0;
	for(n = 1 ; n < mArgc ; n++)
		{
		if(strcmp("-k", mArgv[n]) == 0)
			{
			if(sscanf(mArgv[n+2], "%d/%d", &deltaNum, &deltaDen) != 2)
				{
				deltaNum = DEFAULT_PMODEL_DELTA_NUM;
				deltaDen = DEFAULT_PMODEL_DELTA_DEN;
				}

			cModels[cModel] = CreateCModel(cTemplate[cModel].size,
			  N_SYMBOLS, N_SYMBOLS, deltaNum, deltaDen, PMODEL_MAX_COUNT,
			  hSize, gamma, updateInvComp);
			printf("Creating %llu pModels (tSize: %d, delta = %d/%d)\n",
			  cModels[cModel]->nPModels, cModels[cModel]->ctxSize, deltaNum,
			  deltaDen);
			pModels[cModel] = CreatePModel(N_SYMBOLS);
			cModel++;
			}

		}

	switch(mode)
		{
		case MODEL_STATIC:
			printf("Run mode: MODEL_STATIC\n");
			printf("Model sequence range: %d:%d\n", modelStart, modelEnd);
			printf("Target sequence range: %d:%d\n", targetStart, targetEnd);
			printf("Output sequence range: %d:%d\n", outputStart, outputEnd);
			Run(modelSeq, alphabet, cModels, pModels, cTemplate, nCModels,
			  UPDATE_CMODELS, NULL, modelStart, modelEnd);
			Run(targetSeq, alphabet, cModels, pModels, cTemplate, nCModels,
			  !UPDATE_CMODELS, profile, targetStart, targetEnd);
			break;

		case MODEL_DYNAMIC:
			printf("Run mode: MODEL_DYNAMIC\n");
			printf("Model sequence range: %d:%d\n", modelStart, modelEnd);
			printf("Target sequence range: %d:%d\n", targetStart, targetEnd);
			printf("Output sequence range: %d:%d\n", outputStart, outputEnd);
			Run(modelSeq, alphabet, cModels, pModels, cTemplate, nCModels,
			  UPDATE_CMODELS, NULL, modelStart, modelEnd);
			Run(targetSeq, alphabet, cModels, pModels, cTemplate, nCModels,
			  UPDATE_CMODELS, profile, targetStart, targetEnd);
			break;

		case SELF_MIN:
			printf("Run mode: SELF_MIN\n");
			printf("Target sequence range: %d:%d\n", targetStart, targetEnd);
			printf("Output sequence range: %d:%d\n", outputStart, outputEnd);
			Run(targetSeq, alphabet, cModels, pModels, cTemplate, nCModels,
			  UPDATE_CMODELS, profile, targetStart, targetEnd);
			RevertComplement(targetSeq);
			printf("Current memory allocated by (m/c/re)alloc: %.2f MiB\n",
			  TotalMem() / 1024. / 1024.);
			printf("Reseting cModels...\n");
			for(cModel = 0 ; cModel < nCModels ; cModel++)
				ResetCModel(&cModels[cModel]);

			printf("...done\n");
			printf("Current memory allocated by (m/c/re)alloc: %.2f MiB\n",
			  TotalMem() / 1024. / 1024.);
			if(!(tmpProfile = (float *)Malloc(targetSeq->size * sizeof(float))))
				{
				fprintf(stderr, "Error: out of memory\n");
				return 1;
				}

			Run(targetSeq, alphabet, cModels, pModels, cTemplate, nCModels,
			  UPDATE_CMODELS, tmpProfile, targetSeq->size - targetEnd + 1,
			  targetSeq->size - targetStart + 1);
			ReverseProfile(tmpProfile, targetSeq->size);

			// The final profile is the minimum, point by point, of both runs
			for(n = 0 ; n < targetSeq->size ; n++)
				profile[n] = profile[n] < tmpProfile[n] ?
				  profile[n] : tmpProfile[n];

			break;

		case SELF_LR:
			printf("Run mode: SELF_LR\n");
			printf("Target sequence range: %d:%d\n", targetStart, targetEnd);
			printf("Output sequence range: %d:%d\n", outputStart, outputEnd);
			Run(targetSeq, alphabet, cModels, pModels, cTemplate, nCModels,
			  UPDATE_CMODELS, profile, targetStart, targetEnd);
			break;

		case SELF_RL:
			printf("Run mode: SELF_RL\n");
			printf("Target sequence range: %d:%d\n", targetStart, targetEnd);
			printf("Output sequence range: %d:%d\n", outputStart, outputEnd);
			RevertComplement(targetSeq);
			Run(targetSeq, alphabet, cModels, pModels, cTemplate, nCModels,
			  UPDATE_CMODELS, profile, targetSeq->size - targetEnd + 1,
			  targetSeq->size - targetStart + 1);
			ReverseProfile(profile, targetSeq->size);
			break;

		}

	tac = clock();
	cpuTimeUsed = ((double) (tac - tic)) / CLOCKS_PER_SEC;
    printf("Needed %g s for generating the profile\n", cpuTimeUsed);

	printf("Using a window size of %d bp for low-pass filtering\n", 2 * M + 1);
	if(fpOut)
		{
		if(wigChrom) // Generate a wiggle file format
			{
			fprintf(fpOut, "browser position %s:%d-%d\n", wigChrom,
			  outputStart, outputEnd);
			fprintf(fpOut, "track type=wiggle_0 name=DNAatGlance "
			  "description=\"Information Profile\" visibility=full "
			  "color=0,0,255 viewLimits=0:2 autoScale=off graphType=points "
			  "maxHeightPixels=64:50:11\n");
			fprintf(fpOut, "variableStep chrom=%s\n", wigChrom);
			omitIndexCol = 'n';
			}

		if(omitIndexCol == 'y')
			{
			for(n = outputStart - 1 ; n < outputEnd ; n++)
				if(n >= targetStart - 1 && n < targetEnd && n % S == 0)
					{
					outValue = Filter(profile, targetSeq->size, n, M, w);
					if(outValue == MISSING_BASE)
						fprintf(fpOut, "\n");

					else
						fprintf(fpOut, "%.3f\n", outValue);

					}

			}

		else
			{
			for(n = outputStart - 1 ; n < outputEnd ; n++)
				if(n >= targetStart - 1 && n < targetEnd && n % S == 0)
					{
					outValue = Filter(profile, targetSeq->size, n, M, w);
					if(outValue == MISSING_BASE)
						{
						if(!wigChrom)
							fprintf(fpOut, "\n");

						}

					else
						fprintf(fpOut, "%d\t%.3f\n", n + 1, outValue);

					}

			}

		fclose(fpOut);
		}

	if(plot == 'y')
		{
		if(!(fpPlot = popen("gnuplot -persist", "w")))
			{
			fprintf(stderr, "Error: can't open pipe for gnuplot\n");
			return 1;
			}

		fprintf(fpPlot, "plot '-' with lines\n");
		for(n = outputStart - 1 ; n < outputEnd ; n++)
			if(n >= targetStart - 1 && n < targetEnd && n % S == 0)
				{
				outValue = Filter(profile, targetSeq->size, n, M, w);
				if(outValue == MISSING_BASE)
					fprintf(fpPlot, "\n");

				else
					fprintf(fpPlot, "%d\t%.3f\n", n + 1, outValue);

				}

		fprintf(fpPlot, "e\n");
		fflush(fpPlot);
		getchar();
		fclose(fpPlot);
		}

	cpuTimeUsed = ((double) (tac - clockStart)) / CLOCKS_PER_SEC;
    printf("Total cpu time used: %g s\n", cpuTimeUsed);
	printf("Total memory allocated by (m/c/re)alloc: %.2f MiB\n",
	  TotalMem() / 1024. / 1024.);

	return 0;
	}

